// Copyright 2021 Esri
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import ArcGIS

class ViewController: UIViewController {
    @IBOutlet weak var sceneView: AGSSceneView!

    override func viewDidLoad() {
        super.viewDidLoad()
        setupScene()

        setupUI()

        setupGraphicsOverlay()

        setupAnalysisOverlay()

    }

    // MARK: Scene

    private func setupScene() {
        let scene: AGSScene = {
            let portal = AGSPortal.arcGISOnline(withLoginRequired: false)

            let item = AGSPortalItem(portal: portal, itemID: "7558ee942b2547019f66885c44d4f0b1")

            return AGSScene(item: item)
        }()
        sceneView.scene = scene

        sceneView.touchDelegate = self

    }

    // MARK: UI

    private func setupUI() {
        let clear = UIBarButtonItem(title: "Clear", style: .plain, target: self, action: #selector(clearScene))
        navigationItem.rightBarButtonItem = clear
    }

    @objc
    func clearScene(_ sender: UIBarButtonItem) {
        clearAnalyses()
    }

    // MARK: Graphics

    private let graphicsOverlay = AGSGraphicsOverlay()

    private let observerGraphic: AGSGraphic = {
        let symbol = AGSSimpleMarkerSymbol(style: .circle, color: .black, size: 16)
        symbol.outline = AGSSimpleLineSymbol(style: .solid, color: .white, width: 2)
        return AGSGraphic(geometry: nil, symbol: symbol)
    }()

    @discardableResult
    private func addTargetGraphic(point: AGSPoint) -> AGSGraphic {
        let symbol = AGSSimpleMarkerSymbol(style: .circle, color: .white, size: 16)
        symbol.outline = AGSSimpleLineSymbol(style: .solid, color: .black, width: 2)
        let targetGraphic = AGSGraphic(geometry: point, symbol: symbol, attributes: nil)
        graphicsOverlay.graphics.add(targetGraphic)
        return targetGraphic
    }

    private var targetGraphics: [AGSGraphic] {
        graphicsOverlay.graphics.filter { ($0 as! AGSGraphic) != observerGraphic } as! [AGSGraphic]
    }

    private func setupGraphicsOverlay() {
        graphicsOverlay.graphics.add(observerGraphic)
        graphicsOverlay.sceneProperties?.surfacePlacement = .absolute
        graphicsOverlay.sceneProperties?.altitudeOffset = 5
        sceneView.graphicsOverlays.add(graphicsOverlay)
    }

    // MARK: Line of Sight Analysis

    private let analysisOverlay = AGSAnalysisOverlay()

    private func setupAnalysisOverlay() {
        sceneView.analysisOverlays.add(analysisOverlay)
    }

    private func createLineOfSightAnalysis(observer: AGSGeoElement, target: AGSGeoElement) {
        let line = AGSGeoElementLineOfSight(observerGeoElement: observer, targetGeoElement: target)
        analysisOverlay.analyses.add(line)
    }

    private func clearAnalyses() {
        analysisOverlay.analyses.removeAllObjects()
        observerGraphic.geometry = nil
        graphicsOverlay.graphics.removeObjects(in: targetGraphics)
    }

    private func setObserver(point: AGSPoint) {
        let shouldCreateAnalyses = observerGraphic.geometry == nil
        observerGraphic.geometry = point
        if shouldCreateAnalyses {
            targetGraphics.forEach { targetGraphic in
                createLineOfSightAnalysis(observer: observerGraphic, target: targetGraphic)
            }
        }
    }

    private func clearObserver() {
        analysisOverlay.analyses.removeAllObjects()
        observerGraphic.geometry = nil
    }

    private func addTarget(point: AGSPoint) {
        let targetGraphic = addTargetGraphic(point: point)
        if observerGraphic.geometry != nil {
            createLineOfSightAnalysis(observer: observerGraphic, target: targetGraphic)
        }
    }

}

// MARK: - GeoView Touch Delegate

extension ViewController: AGSGeoViewTouchDelegate {

    func geoView(_ geoView: AGSGeoView, didTapAtScreenPoint screenPoint: CGPoint, mapPoint: AGSPoint) {

        addTarget(point: mapPoint)

    }

    func geoView(_ geoView: AGSGeoView, didLongPressAtScreenPoint screenPoint: CGPoint, mapPoint: AGSPoint) {

        setObserver(point: mapPoint)

    }

    func geoView(_ geoView: AGSGeoView, didMoveLongPressToScreenPoint screenPoint: CGPoint, mapPoint: AGSPoint) {

        setObserver(point: mapPoint)

    }

    func geoView(_ geoView: AGSGeoView, didEndLongPressAtScreenPoint screenPoint: CGPoint, mapPoint: AGSPoint) {

        setObserver(point: mapPoint)

    }

    func geoViewDidCancelLongPress(_ geoView: AGSGeoView) {

        clearObserver()

    }

}


